from . import add
from . import check_migrate
from . import list
from . import migrate
from . import rename
from . import rm
from . import show
from . import update

commands_list = [
    add,
    check_migrate,
    list,
    migrate,
    rename,
    rm,
    show,
    update
]



#!/usr/bin/env python
# coding: utf-8
'Goto - the magic project that takes you where you need to be, now.'
from __future__ import absolute_import, unicode_literals, print_function
from builtins import dict, str  # redefine dict and str to be py3-like in py2.
# http://johnbachman.net/building-a-python-23-compatible-unicode-sandwich.html

import os
import sys
import codecs

from .settings import GOTOPATH
from .gotomagic import text
from .gotomagic.magic import GotoMagic
from .gotomagic.utils import healthcheck, print_utf8, fix_python2, make_sure_we_print_in_utf8
from .commands import commands, usage
from .plugins import plugins



def main():
    fix_python2()
    make_sure_we_print_in_utf8()
    exit_if_unhealthy()
    exit_with_usage_if_needed()

    project = sys.argv[1]
    magic = GotoMagic(project)
    argv = sys.argv[2:]

    commands.update(plugins)

    command, args, options = parse_args(argv, commands.keys())
    exit_if_no_command_and_no_args(command, args)

    output, err = commands[command].run(magic, command, args, options)
    if output:
        print_utf8(output)
    if err:
        print_utf8(err.message)
        exit(1)

    exit(0)


def parse_args(argv, command_names):
    command = None
    for arg in argv:
        if arg in command_names:
            command = arg
            argv.remove(arg)
            break

    args = list(filter(lambda word: not word.startswith('-'), argv))
    options = list(filter(lambda word: word.startswith('-'), argv))

    return command, args, options


def exit_if_unhealthy():
    err = healthcheck()
    if err:
        print_utf8(err.message)
        exit(1)


def exit_with_usage_if_needed():
    if len(sys.argv) < 3:
        print_usage()
        exit(0)


def exit_if_no_command_and_no_args(command, args):
    if not command and len(args) == 0:
        print_usage()
        exit(0)


def print_usage():
    print_utf8(usage())

if __name__ == '__main__':
    main()



# coding: utf-8
from __future__ import unicode_literals

import subprocess
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:40}{1}".format('vscode', 'Opens Visual Studio Code in code folder')


def names():
    return ['vscode', '--vscode']

def run(magic, command, args, options):
    """
    Launch Visual Studio Code in the code folder
    """

    code = magic.get_uri('code')
    if code is None:
        return None, GotoWarning("no_magicword_named_code")

    try:
        subprocess.check_call('code "%s"' % code, shell=True)
    except subprocess.CalledProcessError:
        return None, GotoError("vscode_launch_failed")

    return None, None



# coding: utf-8
from __future__ import unicode_literals
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:10}{1:30}{2}".format('add','<magicword> <url or path>','Add shortcut')

def names():
    return ['add','--add']


def run(magic, command, args, options):
    """
    Add magicword
    """

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword_and_uri", command='add')

    if (len(args) == 1):
        return None, GotoWarning("missing_uri",
                                 magicword=args[0],
                                 command='add')

    magicword = args[0]
    uri = args[1]

    err = magic.add_shortcut(magicword, uri)
    if err:
        return None, err

    err = magic.save()
    if err:
        return None, err

    return 'Added magic word {}'.format(magicword), None



# coding: utf-8
from __future__ import absolute_import, unicode_literals
from builtins import dict, str  # redefine dict and str to be py3-like in py2.


# errors
messages = dict(
    open_failed="""
Oh noes...
    - Something clogged up the magic, and the magicword could not be opened.

    Computer says: {message}
    """,

    magic_could_not_be_saved="""
Oh noes...!

    - Something clogged up the magic, and the magicwords could not be saved.

    Computer says: {message}

    If you want us to fix it - please tell us about this on github here:

        https://github.com/technocake/goto/issues

    Feel free to create a new issue if one does not already exist.

    """,

    subl_launch_failed="""
Error: could not launch Sublime Text

    This is most likely due to not having the subl command in path.

    For more info see:
        http://docs.sublimetext.info/en/latest/command_line/command_line.html
""",

    intellij_launch_failed="""
Error:  could not launch IntelliJ.

    This is most likely due to not having set up the IntelliJ
    Command Launching yet.

    To be able to launch IntelliJ from the command line,
    this feature must be activated from the IntelliJ menu on your system.

    For more info, see:

        https://www.jetbrains.com/help/idea/working-with-the-ide-features-from-command-line.html
    """,

    vscode_launch_failed="""
Error: could not launch vscode

    This is most likely due to not having set up the vscode
    Command Launching yet.

    For more info, see:

        https://code.visualstudio.com/docs/editor/command-line
    """,
)


# coding: utf-8
from __future__ import unicode_literals
from ..gotomagic.text import GotoWarning


def help():
    return "{0:10}{1:30}{2}".format('rename', '<magicword> <new name>', 'Rename shortcut')


def names():
    return ['rename', '--rename']


def run(magic, command, args, options):
    if len(args) == 0:
        return None, GotoWarning("missing_both_magicwords", command=command)

    from_magicword = args[0]

    if len(args) == 1:
        return None, GotoWarning("missing_to_magicword",
                                 command=command, magicword=from_magicword)

    to_magicword = args[1]
    overwrite = '-f' in options or '--force' in options

    err = magic.rename_shortcut(from_magicword, to_magicword, overwrite)
    if err:
        return None, err
    magic.save()

    return "Renamed {} to {}".format(from_magicword, to_magicword), None



# coding: utf-8
from __future__ import unicode_literals


import os
import sys
from ..gotomagic.text import GotoWarning
from .. import settings

# py2-3 cheat sheet: http://python-future.org/compatible_idioms.html
# make Python 2 and 3 use input (instead of raw_input | input)
from builtins import input


def detect_platform():
    """
        Detects if we are on osx, linux or win
    """

    if sys.platform in ['linux', 'linux2']:
        return 'linux'
    elif sys.platform in ['darwin']:
        return 'osx'
    elif sys.platform in ['win32']:
        return 'win'
    raise Exception("Error: Unknown platform. Could not determine if running linux, osx or windows")  # noqa


def is_file(raw_uri):
    ''' checks if the file or folder exist and returns True if so '''
    candidate = os.path.abspath(raw_uri)
    return os.path.exists(candidate)


def create_project_folder(project, scope='private', GOTOPATH=None):
    '''
        Creates project folder in goto state folders,
        if not already existing.

        scope can be either private or shared
    '''
    if GOTOPATH is None:
        GOTOPATH = settings.GOTOPATH

    project_folder = os.path.join(GOTOPATH, 'projects', project)
    scope_folder = os.path.join(GOTOPATH, 'projects', project, scope)

    # Both folders already exist
    if os.path.exists(project_folder)\
    and os.path.exists(scope_folder):
        return

    # project folder exist, BUT it is a file..
    if os.path.isfile(project_folder):
        return

    os.makedirs(scope_folder)


def detect_unescaped_ampersand_url():
    '''
        Detects if user has entered a url with ampersand,
        and accidentaly launched goto in the background,
        cutting the url in half

        based on
        https://stackoverflow.com/questions/24861351/how-to-detect-if-python-script-is-being-run-as-a-background-process
    '''
    if len(sys.argv) < 3:
        return False
    command = sys.argv[2]
    if command in ['add', 'update']:
        try:
            if os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()):
                # Interactive mode
                return False
            else:
                # Background mode
                return True
        # exception triggered if piping output to a file:
        # i.e  goto list > /tmp/lol
        except Exception:
            pass
    return False


def list_jfiles():
    GOTOPATH = settings.GOTOPATH
    files = os.listdir(os.path.join(GOTOPATH, 'projects'))
    jfiles = filter(lambda f: '.json' in f, files)

    return list(jfiles)


def detect_unmigrated_data():
    '''
    Detects if the data in the gotopath folder is
    not migrated to the current latest version.

    The check is to see fi there is any json files
    in the root of ~/.goto/projects
    '''
    return len(list_jfiles()) != 0


def prompt_to_migrate_data():
    ''' Ask the user to migrate their project data '''
    try:
        return 'y' in input('Shall goto migrate your data now? [y|n]: ').lower()  # noqa
    except (IOError, EOFError, RuntimeError, KeyboardInterrupt):
        # In case of Ctrl+D or Ctrl+C
        return False


def handle_unescaped_ampersand_url():
    command = sys.argv[2]
    magicword = sys.argv[3]
    return GotoWarning('unescaped_ampersand_url_detected', command=command, magicword=magicword)  # noqa


def healthcheck():
    '''
    Runs all healthchecks.
        - detect_unescaped_ampersand_url
        - detect_unmigrated_data
    returns GotoWarning if any issues detected.
    '''
    if len(sys.argv) >= 3:
        if detect_unescaped_ampersand_url():
            return handle_unescaped_ampersand_url()

    if detect_unmigrated_data():
        # ugly hack to make it work
        if '--migrate' not in sys.argv and '--check-migrate' not in sys.argv:
            return GotoWarning('data_not_migrated')

    # No issues detected
    return None


# *----------  Python 2 fixes below ---------------------------------* #


def print_utf8(message):
    ''' fixes utf-8 unicode printing bugs in python 2.
        Making sure all printed strings are of the unicode type,
        and that they are encoded to bytes using the utf-8 encoding.

        All internal strings in goto should be utf-8 encoded.

        In python3 this is the default behaviour.
    '''
    if sys.version_info[0] == 2:
        if not isinstance(message, unicode):  # noqa
            message = unicode(message, 'utf-8')  # noqa
        message = message.encode('utf-8')

    print(message)


def fix_python2():
    '''
        Assume all input is utf-8.
        I am sure this will cause issues
    '''
    if sys.version_info[0] == 2:
        sys.argv = map(lambda arg: unicode(arg, 'utf8'), sys.argv)  # noqa


def make_sure_we_print_in_utf8():
    try:
        if sys.stdout.encoding != 'utf-8':
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'strict')
    except:
        pass  # TODO: implement utf-8 encoding of py2.7




# coding: utf-8
from __future__ import unicode_literals


def help():
    return "{0:10}{1:30}{2}".format('list', '[-v]', 'List all shortcuts')


def names():
    return ['list', '--list']


def run(magic, command, args, options):
    """
    List commands
    """
    verbose = '-v' in options or '--verbose' in options
    magicwords = magic.list_shortcuts(verbose=verbose)

    if not magicwords:
        return None, None

    if verbose:
        output = ""
        for word in magicwords:
            uri = magic.get_uri(word)
            if uri is None:
                return None, GotoWarning("magicword_does_not_exist", magicword=word)
            output += "%16s --> %s\n" % (word, uri)
    else:
        output = "\n".join(magicwords)

    return output, None



from unittest import TestCase
import os
import shutil

from ..gotomagic.magic import GotoMagic


# Setting up a temporary gotopath.
TMPGOTOPATH = '/tmp/.goto-unit-tests'
project = '__testgoto__'
tmpgotofile = os.path.join(TMPGOTOPATH, 'projects', project, 'private', 'magicwords.json')  # noqa


class TestMagic(TestCase):
    def setUp(self):
        """ Sets up goto magic to use tmp file """
        self.magic = GotoMagic(project, scope='private', GOTOPATH=TMPGOTOPATH)

    def test_adding_shortcut(self):
        """ Adding shortcuts through the gotomagic module"""
        self.magic.add_shortcut('test', 'http://example.com')
        self.magic.save()

        self.assertTrue(os.path.exists(tmpgotofile),
                        'Adding shortcut on empty project'
                        ' did not create magic file')

    def tearDown(self):
        """ Empty testgotofile after each test """
        if os.path.exists(TMPGOTOPATH):
            # Displaying contents of testfile
            with open(tmpgotofile) as f:
                print(f.read())
            shutil.rmtree(TMPGOTOPATH)



from ..gotomagic.text import GotoError, GotoWarning

def help():
    return "{0:10}{1:30}{2}".format('update', '<magicword> <new url/path>', 'Update shortcut')


def names():
    return [
        'update',
        '--update'
    ]

def run(magic, command, args, options):
    """
    Update magicword
    """

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword_and_uri", command='update')

    if (len(args) == 1):
        return None, GotoWarning("missing_uri",
                                 magicword=args[0],
                                 command='update')

    word = args[0]
    uri = args[1]

    err = magic.update_shortcut(word, uri)
    if err:
        return None, err

    err = magic.save()
    if err:
        return None, err

    return 'Updated magic word %s' % word, None



import setuptools
from setuptools import find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

with open("POST_INSTALL_INSTRUCTIONS.txt", "r") as fh:
    post_install_instructions = fh.read()


setuptools.setup(
    name="magicgoto",
    version="1.6.2",
    author="Robin Aaberg",
    author_email="robin.garen@gmail.com",
    description="Magic goto - goto where you need to be, right now.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gotogoto.ninja",
    include_package_data=True,
    package_dir={
        '': '.'
    },
    packages=find_packages(where='.'),
    python_requires=">=3.3",
    install_requires=[
        'pyperclip',
        'gitpython',
        'future',
    ],
    scripts=[
        'bin/goto',
        'bin/project',
        'bin/_gotoutils',
        'bin/start_goto',
        'bin/install_goto',
    ],
    entry_points={
        'console_scripts': ['the_real_goto.py=goto.the_real_goto:main']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



name='plugins'

from . import copy
from . import intellij
from . import open
from . import subl
from . import vscode
from . import rider

plugins_list = [
    copy,
    intellij,
    open,
    subl,
    vscode,
    rider,
]

plugins = {}

for plugin in plugins_list:
    for name in plugin.names():
        plugins[name] = plugin

__all__ = ['plugins', 'plugins_list']



# coding: utf-8
from __future__ import unicode_literals
import os
import shutil

from ..gotomagic.text import GotoError, GotoWarning
from ..gotomagic.utils import detect_unmigrated_data, prompt_to_migrate_data, list_jfiles, create_project_folder
from .. import settings


def help():
    return "{0:40}{1}".format('--migrate', 'Migrate data to new format')


def names():
    return ['--migrate']


def run(magic, command, args, options):
    """
    migrate magicwords.
    """

    if not detect_unmigrated_data():
        return "Nothing to migrate", None

    if prompt_to_migrate_data():
        return migrate_data()
    else:
        return None, GotoWarning('goto_wont_work_without_migrating_data')


def migrate_data():
    ''' Migrating data

        goto format:
        from v0 - 1.5.x
        to  v1.6.0 and above
    '''
    GOTOPATH = settings.GOTOPATH
    projects_folder = os.path.join(GOTOPATH, 'projects')
    jfiles = list_jfiles()

    migrations = 0
    projects = []
    unmigrated_projects = []

    output = ['Found {} projects to migrate'.format(len(jfiles))]
    warning = None

    for jfile in jfiles:
        project = jfile.split('.json')[0]
        from_file = os.path.join(projects_folder, jfile)
        destination = os.path.join(projects_folder, project, 'private', 'magicwords.json')  # noqa

        # project cmd used to leave empty files for each project
        stubfile = os.path.join(projects_folder, project)
        if os.path.isfile(stubfile):
            os.remove(stubfile)

        create_project_folder(project)

        if move_magicwords(from_file, destination):
            migrations += 1
            projects += [project]
        else:
            unmigrated_projects += [project]
            output.append('Skipping project {} (source file: {} destination file already exists: {}'.format(project, from_file, destination))  # noqa

    output.append('{} of {} projects were migrated'.format(migrations, len(jfiles)))  # noqa
    output.append('Migrated projects: {}'.format(" ".join(projects)))  # noqa

    if not migrations == len(jfiles):
        warning = GotoWarning('not_all_projects_migrated', projects=" ".join(unmigrated_projects))  # noqa

    output = '\n'.join(output)
    return output, warning


def move_magicwords(from_file, destination):
    try:
        if not os.path.exists(destination):
            shutil.move(from_file, destination)
            return True
        else:
            return False
    except shutil.Error:
        # todo: user-friendly warning here.
        raise



# coding: utf-8
from __future__ import unicode_literals

import os
import subprocess
from ..gotomagic.utils import detect_platform
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:10}{1:30}{2}".format('open', '<magicword>', 'Open magicword in window')


def names():
    return ['open', '--open']


def run(magic, command, args, options):
    """
    Open folder
    """
    command = command if command else ''
    verbose = '-v' in options or '--verbose' in options

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword", command)

    magicword = args[0]
    uri = magic.get_uri(magicword)

    if uri is None:
        return None, GotoWarning("magicword_does_not_exist", magicword=magicword)

    uri = os.path.expanduser(uri)
    platform = detect_platform()

    try:
        if platform == 'linux':
            subprocess.call('xdg-open "%s"' % uri, shell=True)
        elif platform == 'osx':
            subprocess.call('open "%s"' % uri, shell=True)
        elif platform == 'win':
            subprocess.call('start "%s"' % uri, shell=True)

    except subprocess.CalledProcessError as e:
        return None, GotoError("open_failed", message=e.message)

    return uri if verbose else None, None



from __future__ import unicode_literals

import subprocess
from ..gotomagic.utils import detect_platform
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:40}{1}".format('idea', 'Opens Rider in code folder')


def names():
    return ['rider', '--rider',]


def run(magic, command, args, options):
    """
    Launch Rider. Open it in the code folder.
    """

    platform = detect_platform()

    if platform in ['osx', 'linux']:
        cmd = "rider"
    elif platform == 'win':
        cmd = "rider.exe"

    code = magic.get_uri('code')
    if code is None:
        return None, GotoWarning("no_magicword_named_code")

    try:
        subprocess.check_call('%s "%s"' % (cmd, code), shell=True)
    except subprocess.CalledProcessError:
        return None, GotoWarning("rider_launch_failed")

    return None, None
